import os
import errno

import numpy as np
import tensorflow as tf
from skimage.external import tifffile as tiff
from tensorflow.python.keras.optimizers import SGD, Adam
from tensorflow.python.keras import backend as K

from deepcell import get_data
from deepcell import make_training_data
from deepcell import rate_scheduler
from deepcell.model_zoo import bn_feature_net_61x61
from deepcell.model_zoo import bn_feature_net_2D
from deepcell.model_zoo import bn_feature_net_skip_2D

from deepcell.training import train_model_conv
from deepcell.training import train_model_sample


RESIZE = True
RESHAPE_SIZE = 256
BINS = 4
EROSION = 0

# filepath constants
DATA_DIR = '/home/foo/test/data'
MODEL_DIR = '/home/foo/test/models'
NPZ_DIR = '/home/foo/test/npz_data'
RESULTS_DIR = '/home/foo/test/results'
EXPORT_DIR = '/home/foo/test/exports'
PREFIX = 'cells/unspecified_nuclear_data/nuclei_broad'
DATA_FILE = 'nuclei_broad_watershed'

# Check for channels_first or channels_last
IS_CHANNELS_FIRST = K.image_data_format() == 'channels_first'
print(IS_CHANNELS_FIRST)

ROW_AXIS = 2 if IS_CHANNELS_FIRST else 1
COL_AXIS = 3 if IS_CHANNELS_FIRST else 2
CHANNEL_AXIS = 1 if IS_CHANNELS_FIRST else -1

WIN_X = 30
WIN_Y = 30

#config = tf.ConfigProto(intra_op_parallelism_threads=22, inter_op_parallelism_threads=22,
#                        allow_soft_placement=True, device_count = {'CPU': 2})
#session = tf.Session(config=config)
#K.set_session(session)

# create these directories if they do not exist
for d in (NPZ_DIR, MODEL_DIR, RESULTS_DIR):
    try:
        os.makedirs(os.path.join(d, PREFIX))
        print('Created new directory:', os.path.join(d, PREFIX))
    except OSError as exc: # Guard against race condition
        if exc.errno != errno.EEXIST:
            raise

# Make the training data for watershed distance transform
make_training_data(
    direc_name=os.path.join(DATA_DIR, PREFIX),
    dimensionality=2,
    file_name_save=os.path.join(NPZ_DIR, PREFIX, DATA_FILE),
    training_direcs=['set1', 'set2'],
    channel_names=['Nuclei'],
    num_of_features=BINS,
    raw_image_direc='processed',
    annotation_direc='annotated',
    reshape_size=RESHAPE_SIZE if RESIZE else None)

if os.path.isfile(os.path.join(NPZ_DIR, PREFIX, DATA_FILE) + '.npz'):
    print('Data Saved to', os.path.join(NPZ_DIR, PREFIX, DATA_FILE) + '.npz')
else:
    raise Exception('Uh Oh!  Your data file did not save properly :(')

# Load the training data from NPZ into a numpy array
training_data = np.load(os.path.join(NPZ_DIR, PREFIX, DATA_FILE + '.npz'))

X, y = training_data['X'], training_data['y']
print('X.shape: {}\ny.shape: {}'.format(X.shape, y.shape))

# save the size of the input data for input_shape model parameter
size = (RESHAPE_SIZE, RESHAPE_SIZE) if RESIZE else X.shape[ROW_AXIS:COL_AXIS + 1]
if IS_CHANNELS_FIRST:
    input_shape = (X.shape[CHANNEL_AXIS], size[0], size[1])
else:
    input_shape = (size[0], size[1], X.shape[CHANNEL_AXIS])

print(input_shape)

# Sample window dimensions
WIN_X = 30
WIN_Y = 30

# Set up other training parameters
n_epoch = 3
batch_size = 10
optimizer = SGD(lr=0.01, decay=1e-6, momentum=0.9, nesterov=True)
lr_sched = rate_scheduler(lr=0.01, decay=0.99)

# First, create a foreground/background separation model

# Instantiate the model
fgbg_model = bn_feature_net_2D(
    receptive_field=WIN_X + WIN_Y + 1,
    n_features=2)

# Train the model
train_model_sample(
    model=fgbg_model,
    dataset=DATA_FILE,
    optimizer=optimizer,
    window_size=(WIN_X, WIN_Y),
    batch_size=batch_size,
    transform=None,  # fg/bg separation
    n_epoch=n_epoch,
    balance_classes=True,
    max_class_samples=None,
    log_dir=os.path.join(DATA_DIR, 'tensorboard_logs'),
    direc_save=os.path.join(MODEL_DIR, PREFIX),
    direc_data=os.path.join(NPZ_DIR, PREFIX),
    expt='sample_fgbg',
    lr_sched=lr_sched,
    rotation_range=180,
    flip=True,
    shear=False)


# Next, Create a model for the watershed energy transform

# Instantiate the model
watershed_model = bn_feature_net_2D(
    receptive_field=WIN_X + WIN_Y + 1,
    n_features=BINS)

# Train the model
train_model_sample(
    model=watershed_model,
    dataset=DATA_FILE,
    optimizer=optimizer,
    batch_size=batch_size,
    n_epoch=n_epoch,
    window_size=(WIN_X, WIN_Y),
    transform='watershed',
    distance_bins=BINS,
    erosion_width=EROSION,
    balance_classes=True,
    max_class_samples=None,
    direc_save=os.path.join(MODEL_DIR, PREFIX),
    direc_data=os.path.join(NPZ_DIR, PREFIX),
    log_dir=os.path.join(DATA_DIR, 'tensorboard_logs'),
    expt='sample_watershed',
    lr_sched=lr_sched,
    rotation_range=180,
    flip=True,
    shear=False)

# Re-initializing dilated models for sample-mode predictions
model_fn = bn_feature_net_61x61
watershed_weights_file = '2018-12-04_nuclei_broad_watershed_sample_watershed.h5'
watershed_weights_file = os.path.join(MODEL_DIR, PREFIX, watershed_weights_file)

fg_bg_weights_file = '2018-12-04_nuclei_broad_watershed_sample_fgbg.h5'
fg_bg_weights_file = os.path.join(MODEL_DIR, PREFIX, fg_bg_weights_file)

run_watershed_model = bn_feature_net_2D(
    receptive_field=WIN_X + WIN_Y + 1,
    dilated=True,
    n_features=BINS,
    input_shape=input_shape)
run_watershed_model.load_weights(watershed_weights_file)

run_fgbg_model = bn_feature_net_2D(
    receptive_field=WIN_X + WIN_Y + 1,
    dilated=True,
    n_features=2,
    input_shape=input_shape)
run_fgbg_model.load_weights(fg_bg_weights_file)

# With the trained model, make predictions on testing data
training_data_file = os.path.join(NPZ_DIR, PREFIX, DATA_FILE + '.npz')

train_dict, test_dict = get_data(training_data_file, mode='conv', seed=21)

X_test, y_test = test_dict['X'], test_dict['y']

test_images = run_watershed_model.predict(X_test)
test_images_fgbg = run_fgbg_model.predict(X_test)

print('watershed transform shape:', test_images.shape)
print('segmentation mask shape:', test_images_fgbg.shape)

argmax_images = []
for i in range(test_images.shape[0]):
    argmax_images.append(np.argmax(test_images[i], axis=CHANNEL_AXIS))
argmax_images = np.array(argmax_images)
argmax_images = np.expand_dims(argmax_images, axis=CHANNEL_AXIS)

print('watershed argmax shape:', argmax_images.shape)


# threshold the foreground/background
# and remove back ground from watershed transform
if IS_CHANNELS_FIRST:
    fg_thresh = test_images_fgbg[:, 1, :, :] > 0.9
else:
    fg_thresh = test_images_fgbg[:, :, :, 1] > 0.9

fg_thresh = np.expand_dims(fg_thresh, axis=CHANNEL_AXIS)
argmax_images_post_fgbg = argmax_images * fg_thresh

# Apply watershed method with the distance transform as seed
from scipy import ndimage
from skimage.morphology import watershed
from skimage.feature import peak_local_max

watershed_images = []
for i in range(argmax_images_post_fgbg.shape[0]):
    if IS_CHANNELS_FIRST:
        image = fg_thresh[i, 0, :, :]
        distance = argmax_images_post_fgbg[i, 0, :, :]
    else:
        image = fg_thresh[i, :, :, 0]
        distance = argmax_images_post_fgbg[i, :, :, 0]

    local_maxi = peak_local_max(test_images[i, :, :, -1], min_distance=15, 
                                exclude_border=False, indices=False, labels=image)

    markers = ndimage.label(local_maxi)[0]
    segments = watershed(-distance, markers, mask=image)
    watershed_images.append(segments)

watershed_images = np.array(watershed_images)
watershed_images = np.expand_dims(watershed_images, axis=CHANNEL_AXIS)

import matplotlib.pyplot as plt

index = 0

fig, axes = plt.subplots(ncols=3, nrows=2, figsize=(15, 15), sharex=True, sharey=True)
ax = axes.ravel()

ax[0].imshow(X_test[index, :, :, 0])
ax[0].set_title('Source Image')

ax[1].imshow(test_images_fgbg[index, :, :, 1])
ax[1].set_title('Segmentation Prediction')

ax[2].imshow(fg_thresh[index, :, :, 0], cmap='jet')
ax[2].set_title('Thresholded Segmentation')

ax[3].imshow(argmax_images[index, :, :, 0], cmap='jet')
ax[3].set_title('Watershed Transform')

ax[4].imshow(argmax_images_post_fgbg[index, :, :, 0], cmap='jet')
ax[4].set_title('Watershed Transform w/o Background')

ax[5].imshow(watershed_images[index, :, :, 0], cmap='jet')
ax[5].set_title('Watershed Segmentation')

fig.tight_layout()
fig.savefig('~/test.png', format='png')
# plt.show()

